/*

  Copyright (c) 1998, Digital Creations, Fredericksburg, VA, USA.  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:
  
    o Redistributions of source code must retain the above copyright
      notice, this list of conditions, and the disclaimer that follows.
  
    o Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions, and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
  
    o Neither the name of Digital Creations nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.
  
  
  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.

  If you have questions regarding this software,
  contact:
 
    Digital Creations L.C.  
    info@digicool.com
 
    (540) 371-6909

*/

#include "Python.h"
#include "ctpublic.h"


static PyObject *ErrorObject;

static PyObject *
SybaseError(char *mess)
{
  PyErr_SetString(ErrorObject, mess);
  return NULL;
}

/* ----------------------------------------------------- */

static void PyVar_Assign(PyObject **v, PyObject *e) { Py_XDECREF(*v); *v=e;}
#define ASSIGN(V,E) PyVar_Assign(&(V),(E))
#define UNLESS(E) if(!(E))
#define UNLESS_ASSIGN(V,E) ASSIGN(V,E); UNLESS(V)
#define OBJECT(O) ((PyObject*)(O))

				
/* Buffer's for CS data types */
#include "sy_occbf.c"

/* Sybase defines */
#include "sy_occdf.c"

/* 
   WARNING

   Callbacks are *really* hard. When they are baked enough to enable,
   the ifdefs will probably go away. Don't try to anable them. The
   ifdefs are really just place holders
   */

/* Context, Connection, and Command declares ==========================
 */

typedef struct {
  PyObject_HEAD
#ifdef CALLBACK_SUPPORT
  PyObject *callbacks;
#endif
  CS_CONTEXT *context;
} Context;

staticforward PyTypeObject ContextType;

typedef struct {
  PyObject_HEAD
#ifdef CALLBACK_SUPPORT
  PyObject *callbacks;
#endif
  CS_CONNECTION *connection;
  Context *context;
} Connection;

staticforward PyTypeObject ConnectionType;

typedef struct {
  PyObject_HEAD
  CS_COMMAND *command;
  Connection *connection;
} Command;

staticforward PyTypeObject CommandType;

#ifdef CALLBACK_SUPPORT
#include "sy_occcb.c"
#endif

/* Client message ===============================================
 */

typedef struct {
  PyObject_HEAD
  CS_CLIENTMSG msg;
} ClientMSG;

staticforward PyTypeObject ClientMSGType;

static ClientMSG *
newClientMSG(CS_CLIENTMSG *msg)
{
  ClientMSG *self;
	
  UNLESS(self = PyObject_NEW(ClientMSG, &ClientMSGType)) return NULL;

  if (msg) memcpy(&self->msg, msg, sizeof(CS_CLIENTMSG));

  return self;
}

static void
ClientMSG_dealloc(ClientMSG *self)
{
  PyMem_DEL(self);
}

static PyObject *
ClientMSG_getattr(ClientMSG *self, char *name)
{
  switch (*name)
    {
    case 's':
      if (strcmp(name, "severity")==0) 
	return PyInt_FromLong((long)self->msg.severity);
      if (strcmp(name, "status")==0) 
	return PyInt_FromLong((long)self->msg.status);
      if (strcmp(name, "sqlstate")==0) 
	return PyString_FromStringAndSize(self->msg.sqlstate, 
					  self->msg.sqlstatelen);      
    case 'm':
      if (strcmp(name, "msgnumber")==0) 
	return PyInt_FromLong((long)self->msg.msgnumber);
      if (strcmp(name, "msgstring")==0) 
	return PyString_FromStringAndSize(self->msg.msgstring, 
					  self->msg.msgstringlen);
    case 'o':
      if (strcmp(name, "osnumber")==0) 
	return PyInt_FromLong((long)self->msg.osnumber);
      if (strcmp(name, "osstring")==0) 
	return PyString_FromStringAndSize(self->msg.osstring, 
					  self->msg.osstringlen);
    case '_':
      if (strcmp(name, "__members__")==0) 
	return Py_BuildValue("sssssss",
			     "severity", "status", "sqlstate", "msgnumber",
			     "msgstring", "osnumber", "osstring" 
			     );
    }

  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
}

static PyTypeObject ClientMSGType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "ClientMSG",			/*tp_name*/
  sizeof(ClientMSG),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)ClientMSG_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)ClientMSG_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Client message data structures"
};

/* Server message ===============================================
 */

typedef struct {
  PyObject_HEAD
  CS_SERVERMSG msg;
} ServerMSG;

staticforward PyTypeObject ServerMSGType;

static ServerMSG *
newServerMSG(CS_SERVERMSG *msg)
{
  ServerMSG *self;
	
  UNLESS(self = PyObject_NEW(ServerMSG, &ServerMSGType)) return NULL;

  if (msg) memcpy(&self->msg, msg, sizeof(CS_SERVERMSG));

  return self;
}

static void
ServerMSG_dealloc(ServerMSG *self)
{
  PyMem_DEL(self);
}

static PyObject *
ServerMSG_getattr(ServerMSG *self, char *name)
{
  switch (*name)
    {
    case 'l':
      if (strcmp(name, "line")==0) 
	return PyInt_FromLong((long)self->msg.line);
    case 'm':
      if (strcmp(name, "msgnumber")==0) 
	return PyInt_FromLong((long)self->msg.msgnumber);
    case 'p':
      if (strcmp(name, "proc")==0) 
	return PyString_FromStringAndSize(self->msg.proc, 
					  self->msg.proclen);      
    case 's':
      if (strcmp(name, "severity")==0) 
	return PyInt_FromLong((long)self->msg.severity);
      if (strcmp(name, "status")==0) 
	return PyInt_FromLong((long)self->msg.status);
      if (strcmp(name, "state")==0) 
	return PyInt_FromLong((long)self->msg.state);
      if (strcmp(name, "sqlstate")==0) 
	return PyString_FromStringAndSize(self->msg.sqlstate, 
					  self->msg.sqlstatelen);      
      if (strcmp(name, "svrname")==0) 
	return PyString_FromStringAndSize(self->msg.svrname, 
					  self->msg.svrnlen);      
    case 't':
      if (strcmp(name, "text")==0) 
	return PyString_FromStringAndSize(self->msg.text, 
					  self->msg.textlen);      

    case '_':
      if (strcmp(name, "__members__")==0) 
	return Py_BuildValue("sssssssss",
			     "line", "msgnumber", "proc", "severity",
			     "status", "state", "sqlstate", "svrname", "text"
			     );
    }

  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
}

static PyTypeObject ServerMSGType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "ServerMSG",			/*tp_name*/
  sizeof(ServerMSG),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)ServerMSG_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)ServerMSG_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Server message data structures"
};

/* Locale ======================================================== 

   Note that this is mostly a place holder for now.

 */

typedef struct {
  PyObject_HEAD
  Context *context;
  CS_LOCALE *locale;
} Locale;

static void
Locale_dealloc(Locale *self)
{
  if (self->locale) 
    {
      Py_BEGIN_ALLOW_THREADS
      cs_loc_drop(self->context->context, self->locale);
      Py_END_ALLOW_THREADS
    }

  Py_DECREF(self->context);
  PyMem_DEL(self);
}

static PyTypeObject LocaleType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "Locale",			/*tp_name*/
  sizeof(Locale),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)Locale_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)0,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Sybase CS_LOCALE"
};

/* DataFMT ======================================================== 
 */

typedef struct {
  PyObject_HEAD
  PyObject *loc;
  CS_DATAFMT fmt;  
} DataFmt;

staticforward PyTypeObject DataFmtType;

static PyObject *
newDataFmt(PyObject *ignored, PyObject *args, PyObject *kw)
{
  DataFmt *self;
  char *name="";
  int namelen=0, datatype=0, format=0, maxlength=0, scale=0, precision=0;
  int status=0, count=0, usertype=0;
  PyObject *loc=NULL;

  static char *kwargs[] = {"name", "datatype", "format", "maxlength",
			   "scale", "precision", "status", "count", 
			   "usertype", "locale"};

  if (!PyArg_ParseTupleAndKeywords(
	  args, kw, "|s#iiiiiiiiO!", kwargs,
	  &name, &namelen, &datatype, &format, &maxlength, &scale, &precision,
	  &status, &count, &usertype, &LocaleType, &loc))
    return NULL; 

  if (namelen > CS_MAX_NAME) return SybaseError("Name too long");

  UNLESS(self = PyObject_NEW(DataFmt, &DataFmtType)) return NULL;
  
  memcpy(self->fmt.name, name, namelen);
  self->fmt.namelen=namelen;
  self->fmt.datatype=datatype;
  self->fmt.format=format;
  self->fmt.maxlength=maxlength;
  self->fmt.scale=scale;
  self->fmt.precision=precision;
  self->fmt.status=status;
  self->fmt.count=count;
  self->fmt.usertype=usertype;
  if (loc)
    self->fmt.locale=((Locale*)loc)->locale;
  else
    {
      self->fmt.locale=NULL;
      loc=Py_None;
    }
  Py_INCREF(loc);
  self->loc=loc;

  return OBJECT(self);
}

static void
DataFmt_dealloc(DataFmt *self)
{
  Py_XDECREF(self->loc);
  PyMem_DEL(self);
}

static PyObject *
DataFmt_getattr(DataFmt *self, char *name)
{

#define INTATTR(N) if (strcmp(name, #N)==0) return PyInt_FromLong((long)self->fmt.N)

  switch (*name) 
    {
    case 'n':
      if (strcmp(name, "name")==0) 
	return PyString_FromStringAndSize(self->fmt.name, self->fmt.namelen);
      break;
    case 'c':
      INTATTR(count);
      break;
    case 'd':
      INTATTR(datatype);
      break;
    case 'f':
      INTATTR(format);
      break;
    case 'm':
      INTATTR(maxlength);
      break;
    case 'p':
      INTATTR(precision);
      break;
    case 's':
      INTATTR(scale);
      INTATTR(status);
      break;
    case 'u':
      INTATTR(usertype);
      break;
    case 'l':
      if (strcmp(name, "locale")==0) 
	{
	  Py_INCREF(self->loc);
	  return self->loc;
	}

    case '_':
      if (strcmp(name, "__members__")==0) 
	return Py_BuildValue("ssssssssss",
			     "name", "count", "datatype", "format",
			     "maxlength", "precision", "scale", "status", 
			     "usertype", "locale"
			     );
    }

#undef INTATTR

  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
}

static int
DataFmt_setattr(DataFmt *self, char *name, PyObject *v)
{

#define INTATTR(N) if (strcmp(name, #N)==0 && PyInt_Check(v)) { \
      self->fmt.N=PyInt_AsLong(v); \
      return 0; }

  if (! v) 
    {
      SybaseError("Cannot delete CS_FORMAT attributes");
      return -1;
    }

  switch (*name) 
    {
    case 'n':
      if (strcmp(name, "name")==0 && PyString_Check(v))
	{
	  int l;

	  l=PyString_GET_SIZE(v);
	  if (l > CS_MAX_NAME) 
	    {
	      SybaseError("Name too long");
	      return -1;
	    }
	  memcpy(self->fmt.name, PyString_AS_STRING(v), l);
	  self->fmt.namelen=l;
	  return 0;
	}
      break;
    case 'c':
      INTATTR(count);
      break;
    case 'd':
      INTATTR(datatype);
      break;
    case 'f':
      INTATTR(format);
      break;
    case 'm':
      INTATTR(maxlength);
      break;
    case 'p':
      INTATTR(precision);
      break;
    case 's':
      INTATTR(scale);
      INTATTR(status);
      break;
    case 'u':
      INTATTR(usertype);
      break;
    case 'l':
      if (strcmp(name, "locale")==0 && v->ob_type==&LocaleType)
	{
	  Py_XDECREF(self->loc);
	  self->loc=v;
	  self->fmt.locale=((Locale*)v)->locale;
	  Py_INCREF(v);
	  return 0;
	}
    }

#undef INTATTR

  PyErr_SetString(PyExc_AttributeError, name);
  return -1;
}

static PyTypeObject DataFmtType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "DataFmt",			/*tp_name*/
  sizeof(DataFmt),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)DataFmt_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)DataFmt_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)DataFmt_setattr,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Sybase CS_DATAFMT data structure"
};

/* Context ================================================================
 */

static PyObject *
newContext(PyObject *ignored, PyObject *args)
{
  Context *self;
  CS_RETCODE r;
  int version=CS_VERSION_100;

  if (args && ! PyArg_ParseTuple(args,"|i", &version)) return NULL;
  
  UNLESS(self = PyObject_NEW(Context, &ContextType)) return NULL;
  self->context=NULL;
#ifdef CALLBACK_SUPPORT
  self->callbacks=NULL;
#endif

  Py_BEGIN_ALLOW_THREADS
  r=cs_ctx_alloc((CS_INT)version, &self->context);
  Py_END_ALLOW_THREADS

  if (r != CS_SUCCEED) 
    {
      SybaseError("Could not connect to Sybase");
      goto err;
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_init(self->context, (CS_INT)version);
  Py_END_ALLOW_THREADS

  if (r  != CS_SUCCEED) 
    {
      SybaseError("Could not initialize Sybase client libraries");
      goto err;
    }

  return OBJECT(self);

err:
  Py_DECREF(self);
  return NULL;
}

static void
Context_dealloc(Context *self)
{
  if (self->context) 
    {
      Py_BEGIN_ALLOW_THREADS
      ct_exit(self->context, CS_FORCE_EXIT);
      cs_ctx_drop(self->context);
      Py_END_ALLOW_THREADS
    }
#ifdef CALLBACK_SUPPORT
  Py_XDECREF(self->callbacks);
#endif

  PyMem_DEL(self);
}

static PyObject *
newConnection(Context *self, PyObject *args)
{
  Connection *c;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args, "")) return NULL;
	
  UNLESS (c = PyObject_NEW(Connection, &ConnectionType)) return NULL;

  c->connection = NULL;

  Py_INCREF(self);
  c->context=self;

  Py_BEGIN_ALLOW_THREADS
  r=ct_con_alloc(self->context, &c->connection);
  Py_END_ALLOW_THREADS

  if (r != CS_SUCCEED)
    {
      Py_DECREF(c);
      return SybaseError("Couldn't allocate a connection");
    }

  return OBJECT(c);
}

static PyObject *
Context_cs_config(Context *self, PyObject *args)
{
  int property, action;
  CS_VOID *buf=NULL;
  CS_INT outlen=0, buflen=CS_UNUSED;
  CS_RETCODE ret;
  PyObject *o=Py_None;

  UNLESS (PyArg_ParseTuple(args,"ii|O:config", &action, &property, &o)) 
    return NULL;

  if ((CS_INT)property == CS_MESSAGE_CB)
    {
#ifdef CALLBACK_SUPPORT
      /* Ugh. A callback. */
      if (action==CS_GET) 
	{
	  if (self->callbacks)
	    {
	      o=PyDict_GetItem(self->callbacks, Py_None);
	      if (! o)
		{
		  PyErr_Clear();
		  o=Py_None;
		}
	    }
	  else o=Py_None;

	  Py_INCREF(o);
	  return o;
	}

      if (action==CS_SET)
	{
	  if (! self->callbacks && Context_allocCallbacks(self) < 0) 
	    return NULL;

	  if (PyDict_SetItem(self->callbacks, Py_None, o) < 0) return NULL;

	  buf=&csmsg_callback;
	}
#else
      return SybaseError("Callbacks are not supported.");
#endif
    }
  else if (action==CS_SET || action==CS_GET)
    {
      
      if (Buffer_Check(o))
	{
	  buf=((Buffer*)o)->val.CHAR;
	  buflen=((Buffer*)o)->ob_size;
	}
      else
	{
	  if (action==CS_SET && PyString_Check(o))
	    {
	      buf=PyString_AS_STRING(o);
	      buflen=PyString_GET_SIZE(o);
	    }
	  else
	    {
	      PyErr_SetString(PyExc_TypeError, "buffer argument expected");
	      return NULL;
	    }
	}
    }

  if (action == CS_SET || action==CS_CLEAR)
    {
      Py_BEGIN_ALLOW_THREADS
      ret=cs_config(self->context, (CS_INT)action, (CS_INT)property,
		    buf, buflen, NULL);
      Py_END_ALLOW_THREADS

      return PyInt_FromLong((long)ret);
    }
 
  Py_BEGIN_ALLOW_THREADS
  ret=cs_config(self->context, (CS_INT)action, (CS_INT)property,
		buf, buflen, &outlen);
  Py_END_ALLOW_THREADS

  return Py_BuildValue("ii", (int)ret, (int)(outlen));  
}

static PyObject *
Context_ct_config(Context *self, PyObject *args)
{
  int property, action;
  CS_VOID *buf=NULL;
  CS_INT outlen=0, buflen=CS_UNUSED;
  CS_RETCODE ret;
  PyObject *o=Py_None;

  UNLESS (PyArg_ParseTuple(args,"ii|O:config", &action, &property, &o)) 
    return NULL;

  if (action==CS_SET || action==CS_GET)
    {
      if (Buffer_Check(o))
	{
	  buf=((Buffer*)o)->val.CHAR;
	  buflen=((Buffer*)o)->ob_size;
	}
      else
	{
	  if (action==CS_SET && PyString_Check(o))
	    {
	      buf=PyString_AS_STRING(o);
	      buflen=PyString_GET_SIZE(o);
	    }
	  else
	    {
	      PyErr_SetString(PyExc_TypeError, "buffer argument expected");
	      return NULL;
	    }
	}
    }

  if (action == CS_SET || action==CS_CLEAR)
    {
      Py_BEGIN_ALLOW_THREADS
      ret=ct_config(self->context, (CS_INT)action, (CS_INT)property,
		    buf, buflen, NULL);
      Py_END_ALLOW_THREADS
      return PyInt_FromLong((long)ret);
    }

  Py_BEGIN_ALLOW_THREADS
  ret=ct_config(self->context, (CS_INT)action, (CS_INT)property,
		buf, buflen, &outlen);
  Py_END_ALLOW_THREADS

  return Py_BuildValue("ii", (int)ret, (int)(outlen));  
}

static PyObject *
Context_diag(Context *self, PyObject *args)
{
  int op, type=CS_UNUSED, index=CS_UNUSED;
  CS_RETCODE r;
  PyObject *o;
  CS_VOID *b=NULL;
  CS_INT ib;
  

  UNLESS (PyArg_ParseTuple(args, "i|ii", &op, &type, &index)) return NULL;

  switch (op)
    {
    case CS_MSGLIMIT:
    case CS_STATUS:
      Py_BEGIN_ALLOW_THREADS
      r=cs_diag(self->context, (CS_INT)op, (CS_INT)type, CS_UNUSED, &ib);
      Py_END_ALLOW_THREADS

      return Py_BuildValue("ii", (int)r, (int)ib);
    case CS_GET:
      switch (type)
	{
	case CS_CLIENTMSG_TYPE:
	  UNLESS (o=(PyObject *)newClientMSG(NULL)) return NULL;
	  b=&((ClientMSG*)o)->msg;
	  break;
	default:
	  return SybaseError("Unsupported context diag type");
	}

      Py_BEGIN_ALLOW_THREADS
      r=cs_diag(self->context, (CS_INT)op, (CS_INT)type, (CS_INT)index, b);
      Py_END_ALLOW_THREADS

      ASSIGN(o, Py_BuildValue("iO", (int)r, o));
      return o;
    }
  
  Py_BEGIN_ALLOW_THREADS
  r=cs_diag(self->context, (CS_INT)op, (CS_INT)type, CS_UNUSED, NULL);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static struct PyMethodDef Context_methods[] = {
  {"ct_con_alloc", (PyCFunction)newConnection, METH_VARARGS,
   "con_alloc() -- allocate a connection object"},
  {"cs_config", (PyCFunction)Context_cs_config, METH_VARARGS,
   "cs_config(action,property[, value]) -- get or set a context property"},
  {"ct_configContext_cs_config", (PyCFunction)Context_ct_config, METH_VARARGS,
   "ct_config(action,property[, value]) -- get or set a context property"},
#ifdef CALLBACK_SUPPORT
  {"ct_callback", (PyCFunction)Context_callback, METH_VARARGS,
   "ct_callback(type[, function] -- Set or get a callback"},
#endif
  {"cs_diag", (PyCFunction)Context_diag, METH_VARARGS,
   "cs_diag(operation, type[, index]) -- Get an error message"},
  {NULL,		NULL}
};

static PyObject *
Context_getattr(Context *self, char *name)
{
  return Py_FindMethod(Context_methods, (PyObject *)self, name);
}

static PyTypeObject ContextType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "Context",			/*tp_name*/
  sizeof(Context),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)Context_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)Context_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Sybase context objects"
};

/* Connection =====================================================
 */

static PyObject *
newCommand(Connection *self, PyObject *args)
{
  Command *c;
  CS_RETCODE r;

  UNLESS (c = PyObject_NEW(Command, &CommandType)) return NULL;
  c->connection=self;
  Py_INCREF(self);
  
  c->command=NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_cmd_alloc(self->connection, &c->command);
  Py_END_ALLOW_THREADS

  if (r != CS_SUCCEED) 
    {
      Py_DECREF(c);
      return SybaseError("Couldn't allocate a command");
    }

  return OBJECT(c);
}

static PyObject *
Connection_capability(Connection *self, PyObject *args)
{
  int type, capability;
  CS_BOOL value;
  CS_INT action;
  PyObject *v;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"ii|O:capability", 
			   &type, &capability, &v)) 
    return NULL;

  if (v)
    {
      if (PyObject_IsTrue(v)) value=CS_TRUE;
      else                    value=CS_FALSE;

      Py_BEGIN_ALLOW_THREADS
      r= ct_capability(self->connection, CS_SET, type, capability, &value);
      Py_END_ALLOW_THREADS

      return PyInt_FromLong((long)r);
    }

  value=CS_FALSE;
  Py_BEGIN_ALLOW_THREADS
  r = ct_capability(self->connection, CS_GET, type, capability, &value);
  Py_END_ALLOW_THREADS

  return Py_BuildValue("ii", (int)r, (int)value);
}

static PyObject *
Connection_con_props(Connection *self, PyObject *args)
{
  int action, property;
  CS_VOID *buf=NULL;
  CS_INT outlen=0, buflen=CS_UNUSED;
  CS_RETCODE r;
  PyObject *o=Py_None;

  UNLESS (PyArg_ParseTuple(args,"iiO:con_props", &action, &property, &o))
    return NULL;
      
  if (Buffer_Check(o))
    {
      buf=((Buffer*)o)->val.CHAR;
      buflen=((Buffer*)o)->ob_size;
    }
  else
    {
      if (action==CS_SET && PyString_Check(o))
	{
	  buf=PyString_AS_STRING(o);
	  buflen=PyString_GET_SIZE(o);
	}
      else
	{
	  PyErr_SetString(PyExc_TypeError, "buffer argument expected");
	  return NULL;
	}
    }

  
  if (action == CS_SET || action==CS_CLEAR)
    {
      Py_BEGIN_ALLOW_THREADS
      r=ct_con_props(self->connection, (CS_INT)action, (CS_INT)property,
		     buf, buflen, NULL);
      Py_END_ALLOW_THREADS
      return PyInt_FromLong((long)r);
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_con_props(self->connection, (CS_INT)action, (CS_INT)property,
		 buf, buflen, &outlen);
  Py_END_ALLOW_THREADS

  return Py_BuildValue("ii", (int)r, (int)(outlen));  
}

static PyObject *
Connection_connect(Connection *self, PyObject *args)
{
  char *server=NULL;
  int l=0;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args, "|s#", &server, &l)) return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_connect(self->connection, server, (CS_INT)l);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Connection_cancel(Connection *self, PyObject *args)
{
  int type;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"i:cancel", &type)) return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_cancel(self->connection, NULL, (CS_INT)type);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Connection_close(Connection *self, PyObject *args)
{
  int option=CS_UNUSED;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"|i", &option)) return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_close(self->connection, (CS_INT)option);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Connection_diag(Connection *self, PyObject *args)
{
  int op, type=CS_UNUSED, index=CS_UNUSED;
  CS_RETCODE r;
  PyObject *o;
  CS_VOID *b=NULL;
  CS_INT ib;
  

  UNLESS (PyArg_ParseTuple(args, "i|ii", &op, &type, &index)) return NULL;

  switch (op)
    {
    case CS_MSGLIMIT:
    case CS_STATUS:
      Py_BEGIN_ALLOW_THREADS
      r=ct_diag(self->connection, (CS_INT)op, (CS_INT)type, CS_UNUSED, &ib);
      Py_END_ALLOW_THREADS

      return Py_BuildValue("ii", (int)r, (int)ib);
    case CS_GET:
      switch (type)
	{
	case CS_CLIENTMSG_TYPE:
	  UNLESS (o=(PyObject *)newClientMSG(NULL)) return NULL;
	  b=&((ClientMSG*)o)->msg;
	  break;
	case CS_SERVERMSG_TYPE:
	  UNLESS (o=(PyObject *)newServerMSG(NULL)) return NULL;
	  b=&((ServerMSG*)o)->msg;
	  break;
	default:
	  return SybaseError("Unsupported connection diag type");
	}

      Py_BEGIN_ALLOW_THREADS
      r=ct_diag(self->connection, (CS_INT)op, (CS_INT)type, (CS_INT)index, b);
      Py_END_ALLOW_THREADS

      ASSIGN(o, Py_BuildValue("iO", (int)r, o));
      return o;
    case CS_EED_CMD:
      return SybaseError("Unsupported connection diag operation");
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_diag(self->connection,(CS_INT)op, (CS_INT)type, CS_UNUSED, NULL);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Connection_options(Connection *self, PyObject *args)
{
  int action, option;
  CS_VOID *buf=NULL;
  CS_INT outlen=0, buflen=CS_UNUSED;
  CS_RETCODE r;
  PyObject *o=Py_None;

  UNLESS (PyArg_ParseTuple(args,"iiO:options", &action, &option, &o))
    return NULL;
      
  if (Buffer_Check(o))
    {
      buf=((Buffer*)o)->val.CHAR;
      buflen=((Buffer*)o)->ob_size;
    }
  else
    {
      if (action==CS_SET && PyString_Check(o))
	{
	  buf=PyString_AS_STRING(o);
	  buflen=PyString_GET_SIZE(o);
	}
      else
	{
	  PyErr_SetString(PyExc_TypeError, "buffer argument expected");
	  return NULL;
	}
    }

  if (action == CS_SET || action==CS_CLEAR)
    {
      Py_BEGIN_ALLOW_THREADS
      r=ct_options(self->connection, (CS_INT)action, (CS_INT)option,
		   buf, buflen, NULL);
      Py_END_ALLOW_THREADS

      return PyInt_FromLong((long)r);
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_options(self->connection, (CS_INT)action, (CS_INT)option,
	       buf, buflen, &outlen);
  Py_END_ALLOW_THREADS

  return Py_BuildValue("ii", (int)r, (int)(outlen));  
}


static struct PyMethodDef Connection_methods[] = {
  {"ct_cmd_alloc", (PyCFunction)newCommand, METH_VARARGS,
   "ct_cmd_alloc() -- allocate a command object"},
#ifdef CALLBACK_SUPPORT
  {"ct_callback", (PyCFunction)Connection_callback, METH_VARARGS,
   "ct_callback(type [, func])"},
#endif
  {"ct_capability", (PyCFunction)Connection_capability, METH_VARARGS,
   "ct_capability(type, capability[, value])"},
  {"ct_con_props", (PyCFunction)Connection_con_props, METH_VARARGS,
   "ct_con_props(action, property[, value])"},
  {"ct_connect", (PyCFunction)Connection_connect, METH_VARARGS,
   "ct_connect(server_name)"},
  {"ct_cancel", (PyCFunction)Connection_cancel, METH_VARARGS,
   "ct_cancel(type)"},
  {"ct_close", (PyCFunction)Connection_close, METH_VARARGS,
   "ct_close([option])"},
  {"ct_diag", (PyCFunction)Connection_diag, METH_VARARGS,
   "ct_diag(operation, type[, index])"},
  {"ct_options", (PyCFunction)Connection_options, METH_VARARGS,
   "ct_options(action, option, buffer)"},
  {NULL,		NULL}		/* sentinel */
};

static void
Connection_dealloc(Connection *self)
{
  if (self->connection) 
    {
      Py_BEGIN_ALLOW_THREADS
      ct_close(self->connection, CS_FORCE_CLOSE);
      ct_con_drop(self->connection);
      Py_END_ALLOW_THREADS
    }
  
  Py_XDECREF(self->context);
  PyMem_DEL(self);
}

static PyObject *
Connection_getattr(Connection *self, char *name)
{
  return Py_FindMethod(Connection_methods, (PyObject *)self, name);
}

static PyTypeObject ConnectionType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "Connection",			/*tp_name*/
  sizeof(Connection),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)Connection_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)Connection_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Sybase database connections"
};

/* Command ================================================================
 */

static PyObject *
Command_command(Command *self, PyObject *args, PyObject *kw)
{
  int type, option=CS_UNUSED, l; 
  char *buffer;
  static char *kwlist[] = {"type", "command", "option", NULL};
  CS_RETCODE r;

  if (!PyArg_ParseTupleAndKeywords(args, kw, "is#|i", kwlist, 
				   &type, &buffer, &l, &option))
             return NULL; 

  Py_BEGIN_ALLOW_THREADS
  r=ct_command(self->command, (CS_INT)type, buffer, (CS_INT)l, (CS_INT)option);
  Py_END_ALLOW_THREADS
 
  return PyInt_FromLong((long)r);
}

static PyObject *
Command_bind(Command *self, PyObject *args)
{
  Buffer *buffer=NULL, *copied=NULL, *indicator=NULL;
  DataFmt *fmt=NULL;
  int item=CS_UNUSED;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args, "|iO!O!O!O!:bind", 
			   &item, 
			   &DataFmtType, &fmt,
			   &BufferType, &buffer, 
			   &BufferType, &copied, 
			   &BufferType, &indicator)) 
    return NULL;

  if (copied && copied->element_type != CS_INT_TYPE)
    return SybaseError("Copied buffer must be CS_INT_TYPE");

  if (indicator && indicator->element_type != CS_SMALLINT_TYPE)
    return SybaseError("Indicator buffer must be CS_SMALLINT_TYPE");
  
  Py_BEGIN_ALLOW_THREADS
  r=ct_bind(self->command, (CS_INT)item, &fmt->fmt,
	    buffer->val.c, 
	    copied ? copied->val.INT : NULL,
	    indicator ? indicator->val.SMALLINT : NULL);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}
				      

static PyObject *
Command_cancel(Command *self, PyObject *args)
{
  int type;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"i:cancel", &type)) return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_cancel(NULL, self->command, (CS_INT)type);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Command_cmd_props(Command *self, PyObject *args)
{
  int action, property;
  CS_VOID *buf=NULL;
  CS_INT outlen=0, buflen=CS_UNUSED;
  CS_RETCODE r;
  PyObject *o=Py_None;

  UNLESS (PyArg_ParseTuple(args,"iiO:cmd_props", &action, &property, &o))
    return NULL;
      
  if (Buffer_Check(o))
    {
      buf=((Buffer*)o)->val.CHAR;
      buflen=((Buffer*)o)->ob_size;
    }
  else
    {
      if (action==CS_SET && PyString_Check(o))
	{
	  buf=PyString_AS_STRING(o);
	  buflen=PyString_GET_SIZE(o);
	}
      else
	{
	  PyErr_SetString(PyExc_TypeError, "buffer argument expected");
	  return NULL;
	}
    }

  if (action == CS_SET || action==CS_CLEAR)
    {
      Py_BEGIN_ALLOW_THREADS
      r=ct_cmd_props(self->command, (CS_INT)action, (CS_INT)property,
		     buf, buflen, NULL);
      Py_END_ALLOW_THREADS

      return PyInt_FromLong((long)r);
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_cmd_props(self->command, (CS_INT)action, (CS_INT)property,
		 buf, buflen, &outlen);
  Py_END_ALLOW_THREADS


  return Py_BuildValue("ii", (int)r, (int)(outlen));  
}

static PyObject *
Command_compute_info(Command *self, PyObject *args)
{
  int type, column;
  Buffer *buffer;
  CS_INT outlen=0;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"iiO!:compute_info", 
			   &type, &column, &BufferType,  &buffer))
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_compute_info(self->command, (CS_INT)type, (CS_INT)column,
		    buffer->val.CHAR, buffer->ob_size, &outlen);
  Py_END_ALLOW_THREADS


  return Py_BuildValue("ii", (int)r, (int)(outlen));  
}

static PyObject *
Command_cursor(Command *self, PyObject *args)
{
  int type, option, namelen, textlen;
  char *name, *text;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"is#s#i!:cursor", 
			   &type, &name, &namelen, &text, &textlen, &option))
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_cursor(self->command, (CS_INT)type,
	      (CS_CHAR *)name, (CS_INT)namelen,
	      (CS_CHAR *)text, (CS_INT)textlen,
	      (CS_INT)option);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Command_describe(Command *self, PyObject *args)
{
  DataFmt *fmt;
  int item;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args, "iO!:describe", &item, &DataFmtType, &fmt))
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_describe(self->command, (CS_INT)item, &fmt->fmt);
  Py_END_ALLOW_THREADS


  return PyInt_FromLong((long)r);
}

static PyObject *
Command_dynamic(Command *self, PyObject *args)
{
  int type, namelen, textlen;
  char *name, *text;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args,"is#s#!:dynamic", 
			   &type, &name, &namelen, &text, &textlen))
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_dynamic(self->command, (CS_INT)type,
	       (CS_CHAR *)name, (CS_INT)namelen,
	       (CS_CHAR *)text, (CS_INT)textlen
	       );
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Command_fetch(Command *self, PyObject *args)
{
  int type=CS_UNUSED, offset=CS_UNUSED, op=CS_UNUSED;
  Buffer *o=NULL;
  CS_RETCODE r;

  UNLESS (PyArg_ParseTuple(args, "|iiiO!", 
			   &type, &offset, &op, &BufferType, &o)) 
    return NULL;

  if (o) 
    {
      if (o->element_type != CS_INT_TYPE || o->ob_size <= 0)
	return SybaseError("Invalid fetch buffer");
      
      Py_BEGIN_ALLOW_THREADS
      r=ct_fetch(self->command, (CS_INT)type, 
		 (CS_INT)offset, (CS_INT)op, &o->val.INT[0]);
      Py_END_ALLOW_THREADS

      return PyInt_FromLong((long)r);
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_fetch(self->command, (CS_INT)type, (CS_INT)offset, (CS_INT)op, NULL);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static PyObject *
Command_get_data(Command *self, PyObject *args)
{
  int i;
  Buffer *buffer;
  CS_RETCODE r;
  CS_INT l;

  UNLESS (PyArg_ParseTuple(args, "iO", &i, &BufferType, &buffer)) return NULL;

  Py_BEGIN_ALLOW_THREADS
  r=ct_get_data(self->command, (CS_INT)i, buffer->val.c, buffer->ob_size, &l);
  Py_END_ALLOW_THREADS

  return Py_BuildValue("ii", (int)r, (int)l);
}

static PyObject *
Command_res_info(Command *self, PyObject *args)
{
  int type;
  CS_INT i, l;
  CS_BOOL b=CS_FALSE;
  CS_USHORT u=0;
  CS_RETCODE r=0;

  UNLESS (PyArg_ParseTuple(args, "i", &type)) return NULL;

  switch (type) 
    {
    case CS_BROWSE_INFO:
      Py_BEGIN_ALLOW_THREADS
      r=ct_res_info(self->command, (CS_INT)type, &b, CS_UNUSED, &l);
      Py_END_ALLOW_THREADS

      return Py_BuildValue("ii", (int)r, (int)b);
    case CS_CMD_NUMBER:
    case CS_NUM_COMPUTES:
    case CS_NUMDATA:
    case CS_NUMORDERCOLS:
    case CS_ROW_COUNT: 
    case CS_TRANS_STATE:
      Py_BEGIN_ALLOW_THREADS
      r=ct_res_info(self->command, (CS_INT)type, &i, CS_UNUSED, &l);
      Py_END_ALLOW_THREADS

      return Py_BuildValue("ii", (int)r, (int)i);
    case CS_MSGTYPE:
      Py_BEGIN_ALLOW_THREADS
      r=ct_res_info(self->command, (CS_INT)type, &u, CS_UNUSED, &l);
      Py_END_ALLOW_THREADS

      return Py_BuildValue("ii", (int)r, (int)u);
    }
  
  return SybaseError("Invalid res_info type");
}

static PyObject *
Command_results(Command *self, PyObject *args)
{
  CS_INT result_type=0;
  CS_RETCODE r;

  if (!PyArg_ParseTuple(args, "")) return NULL; 

  Py_BEGIN_ALLOW_THREADS
  r=ct_results(self->command, &result_type);
  Py_END_ALLOW_THREADS
	
  return Py_BuildValue("ii", (int)r, (int)result_type);
}

static PyObject *
Command_send(Command *self, PyObject *args)
{
  CS_RETCODE r;

  if (!PyArg_ParseTuple(args, "")) return NULL; 
  
  Py_BEGIN_ALLOW_THREADS
  r=ct_send(self->command);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r); 
}

static PyObject *
Command_send_data(Command *self, PyObject *args)
{
  CS_VOID *buf=NULL;
  CS_INT buflen=CS_UNUSED;
  CS_RETCODE r;
  PyObject *o=Py_None;

  UNLESS (PyArg_ParseTuple(args,"O:cmd_props", &o))
    return NULL;
      
  if (Buffer_Check(o))
    {
      buf=((Buffer*)o)->val.CHAR;
      buflen=((Buffer*)o)->ob_size;
    }
  else
    {
      if (PyString_Check(o))
	{
	  buf=PyString_AS_STRING(o);
	  buflen=PyString_GET_SIZE(o);
	}
      else
	{
	  PyErr_SetString(PyExc_TypeError, "buffer argument expected");
	  return NULL;
	}
    }

  Py_BEGIN_ALLOW_THREADS
  r=ct_send_data(self->command, buf, buflen);
  Py_END_ALLOW_THREADS

  return PyInt_FromLong((long)r);
}

static struct PyMethodDef Command_methods[] = {
  {"ct_command", (PyCFunction)Command_command, METH_VARARGS | METH_KEYWORDS,
   "ct_command(type, string [, option]) -- Define a Sybase command"},
  {"ct_bind", (PyCFunction)Command_bind, METH_VARARGS,
   "ct_bind(item, format , buffer[, copied, indicator)"},
  {"ct_cancel", (PyCFunction)Command_cancel, METH_VARARGS,
   "ct_cancel(type) or aCS_COMMAND.cancel([type])"},
  {"ct_cmd_props", (PyCFunction)Command_cmd_props, METH_VARARGS,
   "ct_cmd_props(action, property[, value])"},
  {"ct_compute_info", (PyCFunction)Command_compute_info, METH_VARARGS,
   "ct_compute_info(type, colnum, buffer)"},
  {"ct_cursor", (PyCFunction)Command_cursor, METH_VARARGS,
   "ct_cursor(type, name, text, option)"},
  {"ct_describe", (PyCFunction)Command_describe, METH_VARARGS,
   "ct_describe(item, format)"},
  {"ct_fetch", (PyCFunction)Command_fetch, METH_VARARGS,
   "ct_fetch([type, offset, option, rows_read])"},
  {"ct_get_data", (PyCFunction)Command_get_data, METH_VARARGS,
   "ct_get_data(item, buffer)"},
  {"ct_res_info", (PyCFunction)Command_res_info, METH_VARARGS,
   "ct_res_info(type)"},
  {"ct_results", (PyCFunction)Command_results, METH_VARARGS,
   "ct_results()"},
  {"ct_send_data", (PyCFunction)Command_send_data, METH_VARARGS,
   "ct_send_data(buffer)"},
  {"ct_send", (PyCFunction)Command_send, METH_VARARGS,
   "ct_send() -- Send a Sybase command"},
  {NULL,		NULL}		/* sentinel */
};

static void
Command_dealloc(Command *self)
{
  if (self->command) {
    Py_BEGIN_ALLOW_THREADS
    ct_cmd_drop(self->command);
    Py_END_ALLOW_THREADS
  }
  Py_XDECREF(self->connection);
  PyMem_DEL(self);
}

static PyObject *
Command_getattr(Command *self, char *name)
{
  return Py_FindMethod(Command_methods, (PyObject *)self, name);
}

static PyTypeObject CommandType = {
  PyObject_HEAD_INIT(NULL)
  0,				/*ob_size*/
  "Command",			/*tp_name*/
  sizeof(Command),		/*tp_basicsize*/
  0,				/*tp_itemsize*/
  /* methods */
  (destructor)Command_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)Command_getattr,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  0,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)0,		/*tp_str*/
  (getattrofunc)0,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  /* Space for future expansion */
  0L,0L,
  "Sybase commands"
};


static struct PyMethodDef Module_Level__methods[] = {
  {"cs_ctx_alloc", (PyCFunction)newContext, METH_VARARGS,
   "cs_ctx_alloc() -- Create a Sybase context"},
  {"DATAFMT", (PyCFunction)newDataFmt, METH_VARARGS | METH_KEYWORDS,
   "DATAFMT() -- Create a data format"},
  {"Buffer", (PyCFunction)newBufferobject, METH_VARARGS,
   "Buffer(size) -- Create a new buffer"},
  {NULL, (PyCFunction)NULL, 0, NULL}		/* sentinel */
};

void
initsy_occ()
{
  PyObject *m, *d;
  char *rev="$Revision: 1.3 $";

  /* Create the module and add the functions */
  m = Py_InitModule4("sy_occ", Module_Level__methods,
		     "Bare-metal sybase module\n\n"
		     "$Id: sy_occ.c,v 1.3 1999/11/17 13:49:52 jim Exp $\n" SY_OCCBF_ID SY_OCCDF_ID
		     , (PyObject*)NULL,PYTHON_API_VERSION);

  /* Add some symbolic constants to the module */
  d = PyModule_GetDict(m);

  if (DEFINES(d) < 0) return;

  BufferType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"BufferType", OBJECT(&BufferType));

  ClientMSGType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"CS_CLIENTMSG", OBJECT(&ClientMSGType));

  ServerMSGType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"CS_SERVERMSG", OBJECT(&ServerMSGType));

  DataFmtType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"CS_DATAFMT", OBJECT(&DataFmtType));

  ContextType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"CS_CONTEXT", OBJECT(&ContextType));

  ConnectionType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"CS_CONNECTION", OBJECT(&ConnectionType));

  CommandType.ob_type=&PyType_Type; 
  PyDict_SetItemString(d,"CS_COMMAND", OBJECT(&CommandType));

  ErrorObject = PyString_FromString("sy_occ.error");
  PyDict_SetItemString(d, "error", ErrorObject);

  PyDict_SetItemString(d, "__version__",
		       PyString_FromStringAndSize(rev+11,strlen(rev+11)-2));
  
}
